import datetime as dt
from typing import List, Dict, Optional
import logging
import requests
import ciso8601
import time


API_VERSION = "v1"  # the API access URL is versioned for future compatability and maintenance

# These are all the characters used in elastic search queries, so they should NOT be included in your search str
ALL_RESERVED_CHARS = ['+', '\\', '-', '!', '(', ')', ':', '^', '[', ']', '"', '{', '}', '~', '*', '?', '|', '&', '/']
# However, most query strings are using these characters on purpose, so let's only automatically escape some of them
RARE_RESERVED_CHARS = ['/']

logger = logging.getLogger(__name__)

def sanitize_query(query: str, reserved_char_list: Optional[List[str]] = None) -> str:
    """
    Make sure we properly escape any reserved characters in an elastic search query
    @see https://www.elastic.co/guide/en/elasticsearch/reference/7.17/query-dsl-query-string-query.html#_reserved_characters
    :param query: a full query string
    :param reserved_char_list: characters that need escaping
    :return:
    """
    reserved_chars = reserved_char_list if reserved_char_list else RARE_RESERVED_CHARS
    sanitized = ''
    for char in query:
        if char in reserved_chars:
            sanitized += '\\%s' % char
        else:
            sanitized += char
    return sanitized


def dict_to_list(data: Dict) -> List[Dict]:
    """
    The API returns dicts, but that isn't very restful nor the current standard approach to user-friendly JSON.
    This utility method converts tht into a list of dicts.
    """
    return [{'name': k, 'value': v} for k, v in data.items()]


class MCSearchApiClient:

    API_BASE_URL = "http://ramos.angwin/{}/".format(API_VERSION)
    TIMEOUT_SECS = 30

    # constants used when requesting top terms
    TERM_FIELD_TITLE = "article_title"
    TERM_FIELD_TEXT_CONTENT = "text_content"
    TERM_AGGREGATION_TOP = "top"
    TERM_AGGREGATION_SIGNIFICANT = "significant"
    TERM_AGGREGATION_RARE = "rare"

    def __init__(self, collection: str, api_base_url: str | None = None):
        """
        :param collection: the archive support multiple collections of stories so you have to pass in the
                           name of the collection you want to search against
        :param api_base_url: advanced optional arg: let you override the hard-coded base URL
        """
        self._collection = collection
        if api_base_url:
            self.API_BASE_URL = api_base_url
        self._session = requests.Session()  # better performance to put all HTTP through this one object

    def sample(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs) -> List[Dict]:
        results = self._overview_query(query, start_date, end_date, **kwargs)
        if self._is_no_results(results):
            return []
        return results['matches']

    def top_sources(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs) -> List[Dict]:
        results = self._overview_query(query, start_date, end_date, **kwargs)
        if self._is_no_results(results):
            return []
        return util.dict_to_list(results['topdomains'])

    def top_languages(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs) -> List[Dict]:
        results = self._overview_query(query, start_date, end_date, **kwargs)
        if self._is_no_results(results):
            return []
        return util.dict_to_list(results['toplangs'])

    @staticmethod
    def _is_no_results(results: Dict) -> bool:
        return ('matches' not in results) and ('detail' in results) and (results['detail'] == 'No results found!')

    def count(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs) -> int:
        results = self._overview_query(query, start_date, end_date, **kwargs)
        if self._is_no_results(results):
            return 0
        return results['total']

    def count_over_time(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs) -> List:
        results = self._overview_query(query, start_date, end_date, **kwargs)
        if self._is_no_results(results):
            return []
        data = results['dailycounts']
        to_return = []
        # transform to list of dicts for easier use
        for day_date, day_value in data.items():  # date is in 'YYYY-MM-DD' format
            day = ciso8601.parse_datetime(day_date)
            to_return.append({
                'date': day,
                'timestamp': day.timestamp(),
                'count': day_value,
            })
        return to_return

    @staticmethod
    def _date_query_clause(start_date: dt.datetime, end_date: dt.datetime) -> str:
        return "publication_date:[{} TO {}]".format(start_date.strftime("%Y-%m-%d"), end_date.strftime("%Y-%m-%d"))

    def _overview_query(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs) -> Dict:
        logger.debug("_overview IN: %s %s %s %r", query, start_date, end_date, kwargs)
        # protect query from grabby ANDs
        params = {"q": "({}) AND {}".format(query, self._date_query_clause(start_date, end_date))}
        params.update(kwargs)
        logger.debug("_overview params: %r", params)
        results, _ = self._query("{}/search/overview".format(self._collection), params, method='POST')
        return results

    def article(self, article_id: str) -> Dict:
        # you can extract article_ids from the `article_url` property
        results, _ = self._query("{}/article/{}".format(self._collection, article_id), method='GET')
        return results

    def all_articles(self, query: str, start_date: dt.datetime, end_date: dt.datetime, page_size: int = 1000, **kwargs):
        """
        @return: a generator that yeilds lists of articles, grouped by page.
        """
        params = {"q": "{} AND {}".format(query, self._date_query_clause(start_date, end_date))}
        params.update(kwargs)
        more_pages = True
        next_page_token = None
        while more_pages:
            page, next_page_token = self.paged_articles(query, start_date, end_date, page_size, **kwargs,
                                                        pagination_token=next_page_token)
            if self._is_no_results(page):
                yield []
            else:
                yield page
            # check if there is a link to the next page
            more_pages = False
            if next_page_token:
                more_pages = True

    def paged_articles(self, query: str, start_date: dt.datetime, end_date: dt.datetime,
                       page_size: Optional[int] = 1000,  expanded: bool = False,
                       pagination_token: Optional[str] = None, **kwargs) -> tuple[List[Dict], Optional[str]]:
        """
        @return: one page of stories
        """
        params = {"q": "{} AND {}".format(query, self._date_query_clause(start_date, end_date))}
        if expanded:
            params['expanded'] = 1
        if pagination_token:
            params['resume'] = pagination_token
        params.update(kwargs)
        page, response = self._query("{}/search/result".format(self._collection), params, method='POST')
        if self._is_no_results(page):
            return [], None
        return page, response.headers.get('x-resume-token')

    def terms(self, query: str, start_date: dt.datetime, end_date: dt.datetime, field: str, aggregation: str,
              **kwargs) -> Dict:
        params = {"q": "{} AND {}".format(query, self._date_query_clause(start_date, end_date))}
        params.update(kwargs)
        results, _ = self._query("{}/terms/{}/{}".format(self._collection, field, aggregation), params, method='GET')
        return results

    def _query(self, endpoint: str, params: Dict = None, method: str = 'GET'):
        """
        Centralize making the actual queries here for easy maintenance and testing of HTTP comms
        """
        logger.debug("_query IN: %s ep %s par %s", method, endpoint, params)
        if params and ('domains' in params):  # remove domains param that might be dangling
            del params['domains']
        if params and ('q' in params):
            params['q'] = sanitize_query(params['q'])
        endpoint_url = self.API_BASE_URL+endpoint
        logger.debug("_query SENDING %s ep %s par %r", method, endpoint_url, params)
        start = time.time()
        if method == 'GET':
            r = self._session.get(endpoint_url, params=params, timeout=self.TIMEOUT_SECS)
        elif method == 'POST':
            r = self._session.post(endpoint_url, json=params, timeout=self.TIMEOUT_SECS)
        else:
            raise RuntimeError("Unsupported method of '{}'".format(method))
        duration = time.time() - start

        if r.status_code == 504:
            raise RuntimeError("API returned 504 Gateway Timeout after {} secs. Client timeout: {},"
                                "Endpoint: {}, Params: {}".format(duration, self.TIMEOUT_SECS, endpoint_url, params))

        if r.status_code >= 500:
            raise RuntimeError("API Server Error {}: a bad query string could have triggered this. Endpoint: {},"
                               " Params: {}".format(r.status_code, endpoint_url, params))
        
        return r.json(), r



import collections
import datetime as dt
import importlib.metadata       # to get version for SOFTWARE_ID
import logging
import os
import re
from abc import ABC
from typing import Any, Generator, Iterable, NoReturn, TypeAlias, TypedDict
from operator import itemgetter

from .exceptions import MissingRequiredValue, QueryingEverythingUnsupportedQuery
from .language import terms_without_stopwords

logger = logging.getLogger(__name__)

# helpful for turning any date into the standard Media Cloud date format
MC_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"

DEFAULT_TIMEOUT = 60  # to be used across all the providers; override via one-time call to set_default_timeout

# from api-client/.../api.py
try:
    VERSION = "v" + importlib.metadata.version("mc-providers")
except importlib.metadata.PackageNotFoundError:
    VERSION = "dev"


def set_default_timeout(timeout: int) -> None:
    global DEFAULT_TIMEOUT
    DEFAULT_TIMEOUT = timeout

Item: TypeAlias = dict[str, Any]           # differs between providers?
Items: TypeAlias = list[Item]              # page of items
AllItems: TypeAlias = Generator[Items, None, None]

class Date(TypedDict):
    """
    element of counts list in count_over_time return
    """
    date: dt.date
    timestamp: int
    count: int

class CountOverTime(TypedDict):
    """
    return type for count_over_time
    """
    counts: list[Date]

class Language(TypedDict):
    """
    list element in return value for languages method
    """
    language: str
    value: int
    ratio: float                # really fraction?!

class Source(TypedDict):
    """
    list element in return value for sources method
    """
    source: str
    count: int

class Term(TypedDict):
    """
    list element in return value for words method
    """
    term: str
    count: int
    ratio: float                # really fraction?!

class ContentProvider(ABC):
    """
    An abstract wrapper to be implemented for each platform we want to preview content from.
    Any unimplemented methods raise an Exception
    """
    WORDS_SAMPLE = 500

    LANGUAGE_SAMPLE = 1000

    # default values for _env_val
    # classes which DON'T require a value should define:
    #API_KEY = ""
    #BASE_URL = ""               # subclass can override

    CACHING = 1
    SESSION_ID = ""
    SOFTWARE_ID = f"mc-providers {VERSION}"
    # TIMEOUT *NOT* defined, uses global DEFAULT_TIMEOUT below

    def __init__(self,
                 api_key: str | None = None,
                 base_url: str | None = None,
                 timeout: int | None = None,
                 caching: int | None = None, # handles bool!
                 session_id: str | None = None,
                 software_id: str | None = None):
        """
        api_key and base_url only required by some providers, but accept for all.
        not all providers may use all values, but always accepted to be able
        to detect erroneous args!
        """

        self._api_key = self._env_str(api_key, "API_KEY")
        self._base_url = self._env_str(base_url, "BASE_URL")

        # DEFAULT_TIMEOUT possibly set using set_default_timeout
        self._timeout = self._env_int(timeout, "TIMEOUT", DEFAULT_TIMEOUT)

        if caching == 0:      # not just any falsey value
            # False here, to paraphrase Mitt Romney:
            # yes, my friend, bools are ints
            self._caching = 0
        else:
            self._caching = int(self._env_int(caching, "CACHING"))

        # identify user/session making request (for caching)
        self._session_id = self._env_str(session_id, "SESSION_ID")

        # identify software making request
        # (could be used in User-Agent strings)
        self._software_id = self._env_str(software_id, "SOFTWARE_ID")

    def everything_query(self) -> str:
        raise QueryingEverythingUnsupportedQuery()

    def sample(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 20,
               **kwargs: Any) -> list[dict]:
        raise NotImplementedError("Doesn't support sample content.")

    def count(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> int:
        raise NotImplementedError("Doesn't support total count.")

    def count_over_time(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> CountOverTime:
        raise NotImplementedError("Doesn't support counts over time.")

    def item(self, item_id: str) -> dict:
        raise NotImplementedError("Doesn't support fetching individual content.")

    def words(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 100,
              **kwargs: Any) -> list[Term]:
        raise NotImplementedError("Doesn't support top words.")

    def languages(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> list[Language]:
        raise NotImplementedError("Doesn't support top languages.")

    def sources(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 100,
                **kwargs: Any) -> list[Source]:
        raise NotImplementedError("Doesn't support top sources.")

    def all_items(self, query: str, start_date: dt.datetime, end_date: dt.datetime, page_size: int = 1000,
                  **kwargs: Any) -> AllItems:

        # yields a page of items
        raise NotImplementedError("Doesn't support fetching all matching content.")

    def paged_items(self, query: str, start_date: dt.datetime, end_date: dt.datetime, page_size: int = 1000,
                    **kwargs: Any) -> tuple[list[dict], str | None]:
        # return just one page of items and a pagination token to get next page; implementing subclasses
        # should read in token, offset, or whatever else they need from `kwargs` to determine which page to return
        raise NotImplementedError("Doesn't support fetching all matching content.")

    def normalized_count_over_time(self, query: str, start_date: dt.datetime, end_date: dt.datetime,
                                   **kwargs: Any) -> dict:
        """
        Useful for rendering attention-over-time charts with extra information suitable for normalizing
        HACK: calling _sum_count_by_date for now to solve a problem specific to the Media Cloud provider
        :param query:
        :param start_date:
        :param end_date:
        :param kwargs:
        :return:
        """
        matching_content_counts = self._sum_count_by_date(
            self.count_over_time(query, start_date, end_date, **kwargs)['counts'])
        matching_total = sum([d['count'] for d in matching_content_counts])
        no_query_content_counts = self._sum_count_by_date(
            self.count_over_time(self._everything_query(), start_date, end_date,**kwargs)['counts'])
        no_query_total = sum([d['count'] for d in no_query_content_counts])
        return {
            'counts': _combined_split_and_normalized_counts(matching_content_counts, no_query_content_counts),
            'total': matching_total,
            'normalized_total': no_query_total,
        }

    @classmethod
    def _sum_count_by_date(cls, counts: list[dict]) -> list[Date]:
        """
        Given a list of counts, sum the counts by date
        :param counts:
        :return:
        """
        counts_by_date = collections.defaultdict(int)
        for c in counts:
            date = c['date']
            counts_by_date[date] = 0
            for d in counts:
                if d['date'] == date:
                    counts_by_date[date] += d['count']
        return [{'date': d, 'count': c} for d, c in counts_by_date.items()]

    def _everything_query(self) -> str:
        """
        :return: a query string that can be used to capture matching "everything" 
        """
        return '*'

    # use this if you need to sample some content for top languages
    def _sampled_languages(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 10,
                               **kwargs: Any) -> list[Language]:
        # support sample_size kwarg
        sample_size = kwargs['sample_size'] if 'sample_size' in kwargs else self.LANGUAGE_SAMPLE
        # grab a sample and count terms as we page through it
        sampled_count = 0
        counts: collections.Counter = collections.Counter()
        for page in self.all_items(query, start_date, end_date, limit=sample_size):
            sampled_count += len(page)
            [counts.update(t['language'] for t in page)]
        # clean up results
        results = [Language(language=w, value=c, ratio=c/sampled_count) for w, c in counts.most_common(limit)]
        return results

    def _sample_titles(self, query: str, start_date: dt.datetime, end_date: dt.datetime, sample_size: int,
                             **kwargs: Any) -> Iterable[list[dict[str,str]]]:
        """
        default helper for _sampled_title_words: return a sampling of stories for top words
        """
        # XXX force sort on something non-chronological???
        return self.all_items(query, start_date, end_date, limit=sample_size)

    # use this if you need to sample some content for top words
    def _sampled_title_words(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 100,
                             **kwargs: Any) -> list[Term]:
        # support sample_size kwarg
        sample_size = kwargs.pop('sample_size', self.WORDS_SAMPLE)
        remove_punctuation = bool(kwargs.pop("remove_punctuation", True)) # XXX TEMP?

        # grab a sample and count terms as we page through it
        sampled_count = 0
        counts: collections.Counter = collections.Counter()
        for page in self._sample_titles(query, start_date, end_date, sample_size, **kwargs):
            sampled_count += len(page)
            [counts.update(terms_without_stopwords(t['language'], t['title'], remove_punctuation)) for t in page]
        # clean up results
        results = [Term(term=w, count=c, ratio=c/sampled_count) for w, c in counts.most_common(limit)]
        return results

    # from story-indexer/indexer/story.py:
    # https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    @classmethod
    def _env_var(cls, suffix: str) -> str:
        """
        create CLASS_NAME_SUFFIX name for environment variable from class name and suffix
        """
        name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", cls.__name__)
        return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).upper() + "_" + suffix

    def _missing_value(self, env_var: str) -> NoReturn:
        """
        stand-alone to avoid two copies of the below:

        If PROVIDER_XXX and SOURCE_XXX wanted in exception,
        define PROVIDER_NAME and SOURCE_NAME as class
        members, *BUT* likely means the
        {PROVIDER,SOURCE}_XXX variables in __init__.py would
        need to be defined in this file to avoid a dependency
        loop!  The the _NAME vars existed, they could be
        used to create the PROVIDER_MAP from just a list of
        classes, or a decorator on each class definition!
        """
        raise MissingRequiredValue(type(self).__name__, env_var)

    def _env_str(self, kwval: str | None, variable: str, default: str | None = None) -> str:
        """
        Here when no kwarg passed to constructor.
        variable should be UPPER_SNAKE_CASE string matching the kwarg name!

        NOTE! logic duplicated below in _env_int, so likely that any changes
        neeeded here should be replicated there?!

        MAYBE: prefer env (vs. kwarg likely to have been hard-coded?)
        """
        # 0. if kwarg passed, use it
        if kwval is not None: 
            logger.debug("%r kwval %s '%s'", self, variable, kwval)
            return kwval

        env_var = self._env_var(variable)
        try:
            # 1. Look for OBJ_NAME_VARIABLE env var, returns value if it exits.
            val = os.environ[env_var]
            logger.debug("%r env %s '%s'", self, env_var, val)
            return val
        except KeyError:
            pass

        # 2. If (run-time) default value argument passed, return it.
        if default is not None:
            logger.debug("%r default %s '%s'", self, variable, default)
            return default

        try:
            # 3. Look for class member named "variable", if it exists, return value
            val = getattr(self, variable)
            logger.debug("%r class default %s '%s'", self, variable, val)
            return val
        except AttributeError:
            pass

        # so not "During handling of the above exception"
        self._missing_value(env_var)

    def _env_int(self, kwval: int | None, variable: str, default: int | None = None) -> int:
        """
        ugh: copy of _env_str
        """
        if kwval is not None:
            logger.debug("%r kwval %s %d", self, variable, kwval)
            return kwval

        env_var = self._env_var(variable)
        try:
            # 1. Look for OBJ_NAME_VARIABLE env var, returns value if it exits.
            val = int(os.environ[env_var])
            logger.debug("%r env %s %d", self, env_var, val)
            return val
        except KeyError:
            pass

        # 2. If (run-time) default value argument passed, return it.
        if default is not None:
            logger.debug("%r default %s %d", self, variable, default)
            return default

        try:
            # 3. Look for class member named "variable", if it exists, return value
            val = getattr(self, variable)
            logger.debug("%r class default %s %d", self, variable, val)
            return val
        except AttributeError:
            pass

        # so not "During handling of the above exception"
        self._missing_value(env_var)

def add_missing_dates_to_split_story_counts(counts, start, end, period="day"):
    if start is None and end is None:
        return counts
    new_counts = []
    current = start.date()
    while current <= end.date():
        date_string = current.strftime("%Y-%m-%d %H:%M:%S")
        existing_count = next((r for r in counts if r['date'] == date_string), None)
        if existing_count:
            new_counts.append(existing_count)
        else:
            new_counts.append({'date': date_string, 'count': 0})
        if period == "day":
            current += dt.timedelta(days=1)
        elif period == "month":
            current += dt.timedelta(days=31)
        elif period == "year":
            current += dt.timedelta(days=365)
        else:
            # PB: RuntimeError is for internal Python errors???
            raise RuntimeError("Unsupport time period for filling in missing dates - {}".format(period))
    return new_counts


def _combined_split_and_normalized_counts(matching_results, total_results):
    counts = []
    for day in total_results:
        day_info = {
            'date': day['date'],
            'total_count': day['count']
        }
        matching = [d for d in matching_results if d['date'] == day['date']]
        if len(matching) == 0:
            day_info['count'] = 0
        else:
            day_info['count'] = matching[0]['count']
        if day_info['count'] == 0 or day['count'] == 0:
            day_info['ratio'] = 0
        else:
            day_info['ratio'] = float(day_info['count']) / float(day['count'])
        counts.append(day_info)
    counts = sorted(counts, key=itemgetter('date'))
    return counts



import datetime as dt
import json
import logging
import os
import random
from collections import Counter
from typing import Any, Dict, Iterable, List, Mapping, Optional, Sequence, TypeAlias, TypedDict

# PyPI
import ciso8601
import dateparser     # used for publication_date in IA match_to_row
import numpy as np              # for chunking
from waybacknews.searchapi import SearchApiClient

from .language import stopwords_for_language
from .provider import AllItems, ContentProvider, CountOverTime, Date, Item, Items, Language, Source, Term
from .cache import CachingManager
from .mediacloud import MCSearchApiClient

# don't need a logger per Provider instance
logger = logging.getLogger(__name__)

Counts: TypeAlias = dict[str, int]         # key: count
UrlSearchStrings: TypeAlias = Mapping[str, set[str]] # want Countable[str]

class Overview(TypedDict):
    query: str
    total: int
    topdomains: Counts          # from _format_counts
    toplangs: Counts            # from _format_counts
    dailycounts: Counts         # from _format_day_counts
    matches: list[Dict]         # from _format_match

class OnlineNewsAbstractProvider(ContentProvider):
    """
    All these endpoints accept a `domains: List[str]` search keyword arg.
    and a `base_url: str` constructor keyword arg.
    """

    MAX_QUERY_LENGTH = pow(2, 14)

    # default values for constructor arguments
    API_KEY = ""                # not required

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._client = self.get_client()

    def get_client(self):
        raise NotImplementedError("Abstract provider class should not be implemented directly")

    @classmethod
    def domain_search_string(cls) -> str:
        raise NotImplementedError("Abstract provider class should not be implemented directly")

    def everything_query(self) -> str:
        return '*'

    # Chunk'd
    # NB: it looks like the limit keyword here doesn't ever get passed into the query - something's missing here.
    def sample(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 20,
               **kwargs: Any) -> Items:
        results = []
        for subquery in self._assemble_and_chunk_query_str_kw(query, kwargs):
            this_results = self._client.sample(subquery, start_date, end_date, **kwargs)
            results.extend(this_results)
        
        if len(results) > limit:
            results = random.sample(results, limit)
            
        return self._matches_to_rows(results)

    # Chunk'd
    def count(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> int:
        count = 0
        for subquery in self._assemble_and_chunk_query_str_kw(query, kwargs):
            count += self._client.count(subquery, start_date, end_date, **kwargs)
        return count

    # Chunk'd
    def count_over_time(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> CountOverTime:
        counter: Counter = Counter()
        for subquery in self._assemble_and_chunk_query_str_kw(query, kwargs):
            results = self._client.count_over_time(subquery, start_date, end_date, **kwargs)
            countable = {i['date']: i['count'] for i in results}
            counter += Counter(countable)
        
        counter_dict = dict(counter)
        results = [Date(date=date, timestamp=date.timestamp(), count=count) for date, count in counter_dict.items()]
        # Somehow the order of this list gets out of wack. Sorting before returning for the sake of testability
        results.sort(key=lambda x: x["timestamp"])
        return CountOverTime(counts=results)

    
    @CachingManager.cache()
    def item(self, item_id: str) -> Item:
        one_item = self._client.article(item_id)
        return self._match_to_row(one_item)

    
    # Chunk'd
    def all_items(self, query: str, start_date: dt.datetime, end_date: dt.datetime, page_size: int = 1000, **kwargs: Any) -> AllItems:
        for subquery in self._assemble_and_chunk_query_str(query, **kwargs):
            for page in self._client.all_articles(subquery, start_date, end_date, **kwargs):
                yield self._matches_to_rows(page)

    def paged_items(self, query: str, start_date: dt.datetime, end_date: dt.datetime, page_size: int = 1000, **kwargs: Any)\
            -> tuple[List[Dict], Optional[str]] :
        """
        Note - this is not chunk'd so you can't run giant queries page by page... use `all_items` instead.
        This kwargs should include `pagination_token`, which will get relayed in to the api client and fetch
        the right page of results.
        """
        updated_kwargs = {**kwargs, 'chunk': False}
        query = self._assemble_and_chunk_query_str_kw(query, updated_kwargs)[0]
        page, pagination_token = self._client.paged_articles(query, start_date, end_date, **kwargs)
        return self._matches_to_rows(page), pagination_token

    # Chunk'd
    @CachingManager.cache()
    def words(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 100,
              **kwargs: Any) -> List[Term]:
        chunked_queries = self._assemble_and_chunk_query_str(query, **kwargs)

        # first figure out the dominant languages, so we can remove appropriate stopwords.
        # This method does chunking for you, so just pass the query 
        top_languages = self.languages(query, start_date, end_date, limit=100, **kwargs) 

        represented_languages = [i['language'] for i in top_languages if i['ratio'] > 0.1]
        stopwords = set()
        for lang in represented_languages:
            try:
                stopwords.update(stopwords_for_language(lang))
            except RuntimeError:
                pass  # not stopwords for language, just let them all pass through
            
        # for now just return top terms in article titles
        sample_size = 5000
        
        # An accumulator for the subqueries
        results_counter: Counter = Counter({})
        for subquery in chunked_queries:
            this_results = self._client.terms(subquery, start_date, end_date,
                                     self._client.TERM_FIELD_TITLE, self._client.TERM_AGGREGATION_TOP)
            
            if "detail" not in this_results:
                results_counter += Counter(this_results)
        
        results = dict(results_counter)
            
        # and clean up results to return
        top_terms = [Term(term=t.lower(), count=c, ratio=c/sample_size) for t, c in results.items()
                     if t.lower() not in stopwords]
        top_terms = sorted(top_terms, key=lambda x:x["count"], reverse=True)
        return top_terms

    # Chunk'd
    def languages(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 10,
                  **kwargs: Any) -> List[Language]:
        
        matching_count = self.count(query, start_date, end_date, **kwargs)

        results_counter: Counter = Counter({})
        for subquery in self._assemble_and_chunk_query_str_kw(query, kwargs):
            this_languages = self._client.top_languages(subquery, start_date, end_date, **kwargs)
            countable = {item["name"]: item["value"] for item in this_languages}
            results_counter += Counter(countable)
        
        top_languages = [Language(language=name, value=value, ratio=value/matching_count) for name, value in results_counter.items()]
        
        # Sort by count
        top_languages = sorted(top_languages, key=lambda x: x['value'], reverse=True)
        return top_languages[:limit]

    # Chunk'd
    def sources(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 100,
                **kwargs: Any) -> List[Source]:
        
        results_counter: Counter = Counter({})
        for subquery in self._assemble_and_chunk_query_str_kw(query, kwargs):
            results = self._client.top_sources(subquery, start_date, end_date)
            countable = {source['name']: source['value'] for source in results}
            results_counter += Counter(countable)
        
        cleaned_sources = [Source(source=source, count=count) for source, count in results_counter.items()]
        cleaned_sources = sorted(cleaned_sources, key=lambda x: x['count'], reverse=True)
        return cleaned_sources

    @classmethod
    def _assemble_and_chunk_query_str(cls, base_query: str, chunk: bool = True, **kwargs: Any) -> list[str]:
        """
        If a query string is too long, we can attempt to run it anyway by splitting the domain substring (which is
        guaranteed to be only a sequence of ANDs) into parts, to produce multiple smaller queries which are collectively
        equivalent to the original.

        Because we have this chunking thing implemented, and the filter behavior never interacts with the domain search
        behavior, we can just put the two different search fields into two different sets of behavior at the top.
        There's obvious room to optimize, but this gets the done job.
        """
        logger.debug("AP._assemble_and_chunk_query_str %s %s %r", base_query, chunk, kwargs)
        domains = kwargs.get('domains', [])

        filters = kwargs.get('filters', [])

        if chunk and (len(base_query) > cls.MAX_QUERY_LENGTH):
            # of course there still is the possibility that the base query is too large, which
            # cannot be fixed by this method
            raise RuntimeError(f"Base Query cannot exceed {cls.MAX_QUERY_LENGTH} characters")

        # Get Domain Queries
        domain_queries = []
        if len(domains) > 0:
            domain_queries = [cls._assembled_query_str(base_query, domains=domains)]
            domain_queries_too_big = any([len(q_) > cls.MAX_QUERY_LENGTH for q_ in domain_queries])

            domain_divisor = 2

            if chunk and domain_queries_too_big:
                while domain_queries_too_big:
                    chunked_domains = np.array_split(domains, domain_divisor)
                    domain_queries = [cls._assembled_query_str(base_query, domains=dom) for dom in chunked_domains]
                    domain_queries_too_big = any([len(q_) > cls.MAX_QUERY_LENGTH for q_ in domain_queries])
                    domain_divisor *= 2
                
        # Then Get Filter Queries
        filter_queries = []
        if len(filters) > 0:
            filter_queries = [cls._assembled_query_str(base_query, filters=filters)]
            filter_queries_too_big = any([len(q_) > cls.MAX_QUERY_LENGTH for q_ in filter_queries])

            filter_divisor = 2
            if chunk and filter_queries_too_big:
                while filter_queries_too_big:
                    chunked_filters = np.array_split(filters, filter_divisor)
                    filter_queries = [cls._assembled_query_str(base_query, filters=filt) for filt in chunked_filters]
                    filter_queries_too_big = any([len(q_) > cls.MAX_QUERY_LENGTH for q_ in filter_queries])
                    filter_divisor *= 2
            
        # There's a (probably not uncommon) edge case where we're searching against no collections at all,
        # so just do it manually here.
        if len(domain_queries) == 0 and len(filter_queries) == 0:
            queries = [cls._assembled_query_str(base_query)]
        
        else:
            queries = domain_queries + filter_queries
        
        return queries

    @staticmethod
    def _prune_kwargs(kwargs: dict[str, Any]) -> None:
        """
        takes a query **kwargs dict and removes keys that
        are processed in this library, and should not be passed to clients.
        """
        kwargs.pop("chunk", None) # bool
        kwargs.pop("domains", None) # Iterable[str]
        kwargs.pop("filters", None) # Iterable[str]
        kwargs.pop("url_search_strings", None) # dict[str, Iterable[str]]
        kwargs.pop("url_search_string_domain", None) # bool: TEMP

    @classmethod
    def _check_kwargs(cls, kwargs: dict[str, Any]) -> None:
        """
        check for unknown/misspelled kwargs

        called with kwargs dict after query arguments removed
        copyies kwargs, removes local-only keys and raises
        exception if anything remains
        """
        kwcopy = kwargs.copy()
        cls._prune_kwargs(kwcopy)
        if kwcopy:
            exstring = ", ".join(kwcopy) # join key names
            raise TypeError(f"unknown keyword args: {exstring}")

    @classmethod
    def _assemble_and_chunk_query_str_kw(cls, base_query: str, kwargs: dict = {}) -> list[str]:
        """
        takes kwargs as *dict*, removes items that shouldn't be sent to _client
        """
        chunk = kwargs.pop("chunk", True)
        queries = cls._assemble_and_chunk_query_str(base_query, chunk=chunk, **kwargs)
        cls._prune_kwargs(kwargs)
        return queries

    @classmethod
    def _selector_query_clauses(cls, kwargs: dict) -> list[str]:
        """
        take domains and filters kwargs and
        returns a list of query_strings to be OR'ed together
        (to be AND'ed with user query *or* used as a filter)
        """
        logger.debug("AP._selector_query_clauses IN: %r", kwargs)
        selector_clauses = []

        domains = kwargs.get('domains', [])
        if len(domains) > 0:
            domain_strings = " OR ".join(domains)
            selector_clauses.append(f"{cls.domain_search_string()}:({domain_strings})")
            
        # put all filters in single query string (NOTE: additive)
        filters = kwargs.get('filters', [])
        if len(filters) > 0:
            for filter in filters:
                if "AND" in filter:
                    # parenthesize if any chance it has a grabby AND.
                    selector_clauses.append(f"({filter})")
                else:
                    selector_clauses.append(filter)
        logger.debug("AP._selector_query_clauses OUT: %s", selector_clauses)
        return selector_clauses

    @classmethod
    def _selector_count(cls, kwargs: dict) -> int:
        return len(kwargs.get('domains', [])) + len(kwargs.get('filters', []))

    @classmethod
    def _selector_query_string_from_clauses(cls, clauses: list[str]) -> str:
        return " OR ".join(clauses)

    @classmethod
    def _selector_query_string(cls, kwargs: dict) -> str:
        """
        takes kwargs (as dict) return a query_string to be AND'ed with
        user query or used as a filter.
        """
        return cls._selector_query_string_from_clauses(cls._selector_query_clauses(kwargs))

    @classmethod
    def _assembled_query_str(cls, query: str, **kwargs: Any) -> str:
        logger.debug("_assembled_query_str IN: %s %r", query, kwargs)
        sqs = cls._selector_query_string(kwargs) # takes dict
        if sqs:
            q = f"({query}) AND ({sqs})"
        else:
            q = query
        logger.debug("_assembled_query_str OUT: %s", q)
        return q

    @classmethod
    def _matches_to_rows(cls, matches: List) -> Items:
        raise NotImplementedError()

    @classmethod
    def _match_to_row(cls, match: Dict) -> Item:
        raise NotImplementedError()

    def __repr__(self) -> str:
        # important to keep this unique among platforms so that the caching works right
        return "OnlineNewsAbstractProvider"


class OnlineNewsWaybackMachineProvider(OnlineNewsAbstractProvider):
    """
    All these endpoints accept a `domains: List[str]` keyword arg.
    """
    BASE_URL = ""               # SearchApiClient has default

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)  # will call get_client

    def get_client(self):
        client = SearchApiClient("mediacloud", self._base_url)
        if self._timeout:
            client.TIMEOUT_SECS = self._timeout
        return client

    @classmethod
    def domain_search_string(cls) -> str:
        return "domain"

    @classmethod
    def _matches_to_rows(cls, matches: List) -> Items:
        return [OnlineNewsWaybackMachineProvider._match_to_row(m) for m in matches]

    @classmethod
    def _match_to_row(cls, match: Dict) -> Dict:
        return {
            'media_name': match['domain'],
            'media_url': "http://"+match['domain'],
            'id': match['archive_playback_url'].split("/")[4],  # grabs a unique id off archive.org URL
            'title': match['title'],
            'publish_date': dateparser.parse(match['publication_date']),
            'url': match['url'],
            'language': match['language'],
            'archived_url': match['archive_playback_url'],
            'article_url': match['article_url'],
        }

    def __repr__(self) -> str:
        # important to keep this unique among platforms so that the caching works right
        return "OnlineNewsWaybackMachineProvider"


################
# helpers for formatting url_search_strings (only enabled for MC)
# the helpers are only needed because of the TEMP url_search_string_domain

def format_and_append_uss(uss: str, url_list: list[str]) -> None:
    """
    The ONE place that knows how to format a url_search_string!!!
    (ie; what to put before and after one).

    Appends to `url_list` argument!

    NOTE! generates "unsanitized" (unsanitary?) strings!!

    Currently (11/2024) A URL Search String should:
    1. Start with fully qualified domain name WITHOUT http:// or https://
    2. End with "*"
    """
    # currently url_search_strings MUST start with fully
    # qualified domain name (FQDN) without scheme or
    # leading slashes, and MUST end with a *!
    if not uss.endswith("*"):
        uss += "*"
    url_list.append(f"http\\://{uss}")
    url_list.append(f"https\\://{uss}")

def match_formatted_search_strings(fuss: list[str]) -> str:
    """
    takes list of url search_string formatted by `format_and_append_uss`
    returns query_string fragment
    """
    assert fuss
    urls_str = " OR ".join(fuss)
    return f"url:({urls_str})"


class OnlineNewsMediaCloudProvider(OnlineNewsAbstractProvider):
    """
    Provider interface to access new mediacloud-news-search archive. 
    All these endpoints accept a `domains: List[str]` keyword arg.
    """
    BASE_URL = "http://ramos.angwin:8000/v1/"
    INDEX_PREFIX = "mc_search"

    def __init__(self, **kwargs: Any):
        # maybe take comma separated list?
        self._index = self._env_str(kwargs.pop("index_prefix", None), "INDEX_PREFIX") + "-*"
        super().__init__(**kwargs)

    def get_client(self):
        api_client = MCSearchApiClient(collection=self._index, api_base_url=self._base_url)
        if self._timeout:
            api_client.TIMEOUT_SECS = self._timeout
        return api_client

    @classmethod
    def domain_search_string(cls) -> str:
        return "canonical_domain"

    @classmethod
    def _matches_to_rows(cls, matches: List) -> List:
        return [cls._match_to_row(m) for m in matches]

    @staticmethod
    def _match_to_row(match: Dict) -> Dict:
        story_info = {
            'id': match['id'],
            'media_name': match['canonical_domain'],
            'media_url': match['canonical_domain'],
            'title': match['article_title'],
            'publish_date': dt.date.fromisoformat(match['publication_date']),
            'url': match['url'],
            'language': match['language'],
            'indexed_date': ciso8601.parse_datetime(match['indexed_date']+"Z"),
        }
        if 'text_content' in match:
            story_info['text'] = match['text_content']
        return story_info

    def __repr__(self) -> str:
        return "OnlineNewsMediaCloudProvider"

    def _is_no_results(self, results: Overview) -> bool:
        """
        used to test _overview_query results
        """
        return self._client._is_no_results(results)

    @classmethod
    def _selector_query_clauses(cls, kwargs: dict) -> list[str]:
        """
        take domains, filters, url_search_strings as kwargs
        return a list of query_strings to be OR'ed together
        (to be AND'ed with user query or used as a filter)
        """
        logger.debug("MC._selector_query_clauses IN: %r", kwargs)
        selector_clauses = super()._selector_query_clauses(kwargs)

        # Here to try to get web-search out of query
        # formatting biz.  Accepts a Mapping indexed by
        # domain_string, of lists (or sets!) of search_strings.
        url_search_strings: UrlSearchStrings = kwargs.get('url_search_strings', {})
        if url_search_strings:
            # Unclear if domain field check actually helps at all,
            # so make it optional for testing.
            if kwargs.get("url_search_string_domain", True): # TEMP: include canonincal_domain:
                domain_field = cls.domain_search_string()

                # here with mapping of cdom => iterable[search_string]
                for cdom, search_strings in url_search_strings.items():
                    fuss: List[str] = [] # formatted url_search_strings
                    for sstr in search_strings:
                        format_and_append_uss(sstr, fuss)

                    mfuss = match_formatted_search_strings(fuss)
                    selector_clauses.append(
                        f"({domain_field}:{cdom} AND {mfuss})")

                    format_and_append_uss(cdom, fuss)
            else: # make query without domain (name) field check
                # collect all the URL search strings
                fuss = []
                for cdom, search_strings in url_search_strings.items():
                    for sstr in search_strings:
                        format_and_append_uss(sstr, fuss)

                # check all the urls in one swell foop!
                selector_clauses.append(
                        match_formatted_search_strings(fuss))

        logger.debug("MC._selector_query_clauses OUT: %s", selector_clauses)
        return selector_clauses

    @classmethod
    def _selector_count(cls, kwargs: dict) -> int:
        url_search_strings: UrlSearchStrings = kwargs.get('url_search_strings', {})
        count = super()._selector_count(kwargs)
        if url_search_strings:
            count += sum(map(len, url_search_strings.values()))
        return count

    def count(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> int:
        logger.debug("MC.count %s %s %s %r", query, start_date, end_date, kwargs)
        # no chunking on MC
        results = self._overview_query(query, start_date, end_date, **kwargs)
        return self._count_from_overview(results)

    def _count_from_overview(self, results: Overview) -> int:
        """
        used in .count() and .languages()
        """
        if self._is_no_results(results):
            logger.debug("MC._count_from_overview: no results")
            return 0
        count = results['total']
        logger.debug("MC._count_from_overview: %s", count)
        return count

    def count_over_time(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> CountOverTime:
        logger.debug("MC.count_over_time %s %s %s %r", query, start_date, end_date, kwargs)
        results = self._overview_query(query, start_date, end_date, **kwargs)
        to_return: List[Date] = []
        if not self._is_no_results(results):
            data = results['dailycounts']
            # transform to list of dicts for easier use: process in sorted order
            for day_date in sorted(data):  # date is in 'YYYY-MM-DD' format
                dt = ciso8601.parse_datetime(day_date) # PB: is datetime!!
                to_return.append(Date(
                    date=dt.date(), # PB: was returning datetime!
                    timestamp=int(dt.timestamp()), # PB: conversion may be to local time!!
                    count=data[day_date]
                ))
        logger.debug("MC.count_over_time %d items", len(to_return))
        return CountOverTime(counts=to_return)

    # NB: limit argument ignored, but included to keep mypy quiet
    def sample(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 20, **kwargs: Any) -> List[Dict]:
        logger.debug("MC.sample %s %s %s %r", query, start_date, end_date, kwargs)
        results = self._overview_query(query, start_date, end_date, **kwargs)
        if self._is_no_results(results):
            rows = []
        else:
            rows = self._matches_to_rows(results['matches'])
        logger.debug("MC.sample: %d rows", len(rows))
        return rows             # could slice w/ [:limit]!

    def languages(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 10,
                  **kwargs: Any) -> List[Language]:
        logger.debug("MC.languages %s %s %s %r", query, start_date, end_date, kwargs)
        kwargs.pop("sample_size", None)
        results = self._overview_query(query, start_date, end_date, **kwargs)
        if self._is_no_results(results):
            return []
        matches = self._count_from_overview(results)
        top_languages = [Language(language=name, value=value, ratio=value/matches)
                         for name, value in results['toplangs'].items()]
        logger.debug("MC.languages: _overview returned %d items", len(top_languages))

        # Sort by count
        top_languages = sorted(top_languages, key=lambda x: x['value'], reverse=True)
        items = top_languages[:limit]

        logger.debug("MC.languages: returning %d items", len(items))
        return items

    def sources(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 100,
                **kwargs: Any) -> List[Source]:
        logger.debug("MC.sources %s %s %s %r", query, start_date, end_date, kwargs)
        results = self._overview_query(query, start_date, end_date, **kwargs)
        items: list[Source]
        if self._is_no_results(results):
            items = []
        else:
            cleaned_sources = [Source(source=source, count=count) for source, count in results['topdomains'].items()]
            items = sorted(cleaned_sources, key=lambda x: x['count'], reverse=True)
        logger.debug("MC.sources: %d items", len(items))
        return items

    @CachingManager.cache('overview')
    def _overview_query(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> Overview:
        logger.debug("MC._overview %s %s %s %r", query, start_date, end_date, kwargs)

        # no chunking on MC
        q = self._assembled_query_str(query, **kwargs)
        self._prune_kwargs(kwargs)
        return self._client._overview_query(q, start_date, end_date, **kwargs)

    @classmethod
    def _assemble_and_chunk_query_str(cls, base_query: str, chunk: bool = True, **kwargs: Any) -> list[str]:
        """
        Called by OnlineNewsAbstractProvider.all_items, .words;
        ignores chunking!
        """
        logger.debug("MC._assemble_and_chunk_query_str %s %s %r", base_query, chunk, kwargs)
        return [cls._assembled_query_str(base_query, **kwargs)]

################################################################
# here with original code dragged up from mediacloud.py and news-search-api/api.py

# imports here in case split out into its own file
import base64
import json
import os
import time
from typing import Callable, TypeAlias

import elasticsearch
import mcmetadata.urls as urls
from elasticsearch_dsl import Search, Response
from elasticsearch_dsl.document_base import InstrumentedField
from elasticsearch_dsl.function import RandomScore
from elasticsearch_dsl.query import FunctionScore, Match, Range, Query, QueryString
from elasticsearch_dsl.response import Hit
from elasticsearch_dsl.utils import AttrDict

from .language import terms_without_stopwords

Field: TypeAlias = str | InstrumentedField # quiet mypy complaints
FilterTuple: TypeAlias = tuple[int, Query | None]

_ES_MAXPAGE = 1000              # define globally (ie; in .providers)???

# return value for _overview
_EMPTY_OVERVIEW = Overview(query="", total=-1, topdomains={}, toplangs={}, dailycounts={}, matches=[])

# Was publication_date, but web-search always passes indexed_date.
# identical indexed_date values (without fractional seconds?!)  have
# been seen in the wild (entire day 2024-01-10).
_DEF_SORT_FIELD = "indexed_date"
_DEF_SORT_ORDER = "desc"

# Secondary sort key to break ties if sort_field is non-empty.
# Used as sole sort key if "sort_field" argument or
# _DEF_SORT_FIELD (above) is empty string.
# _doc is documented as most efficient sort key at:
# https://www.elastic.co/guide/en/elasticsearch/reference/current/sort-search-results.html
#
# But at
# https://www.elastic.co/guide/en/elasticsearch/reference/current/paginate-search-results.html
#   "Elasticsearch uses Lucene’s internal doc IDs as tie-breakers. These
#   internal doc IDs can be completely different across replicas of the
#   same data. When paging search hits, you might occasionally see that
#   documents with the same sort values are not ordered consistently."
#
# HOWEVER: use of session_id/preference should route all requests
# from the same session to the same shards for each successive query.
_SECONDARY_SORT_ARGS = {"_doc": "asc"}

def _sanitize(s: str) -> str:
    """
    quote slashes to avoid interpretation as /regexp/
    as done by _sanitize_es_query in mc_providers/mediacloud.py client library

    should only be used by SanitizedQueryString!
    """
    return s.replace("/", r"\/")

class SanitizedQueryString(QueryString):
    """
    query string (expression) with quoting
    """
    def __init__(self, query: str, **kwargs: Any):
        super().__init__(query=_sanitize(query), **kwargs)

def _format_match(hit: Hit, expanded: bool = False) -> dict:
    """
    from news-search-api/client.py EsClientWrapper.format_match
    """
    res = {
        "article_title": getattr(hit, "article_title", None),
        "publication_date": getattr(hit, "publication_date", "")[:10] or None,
        "indexed_date": getattr(hit, "indexed_date", None),
        "language": getattr(hit, "language", None),
        "full_langauge": getattr(hit, "full_language", None),
        "url": getattr(hit, "url", None),
        "original_url": getattr(hit, "original_url", None),
        "canonical_domain": getattr(hit, "canonical_domain", None),
        "id": hit.meta.id        # PB: was re-hash of url!
    }
    if expanded:
        res["text_content"] = getattr(hit, "text_content", None)
    return res

def _format_day_counts(bucket: list) -> Counts:
    """
    from news-search-api/client.py EsClientWrapper.format_count

    used to format "dailycounts" aggregation result

    takes [{"key_as_string": "YYYY-MM-DDT00:00:00.000Z", "doc_count": count}, ....]
    and returns {"YYYY-MM-DD": count, ....}
    """
    return {item["key_as_string"][:10]: item["doc_count"] for item in bucket}

def _format_counts(bucket: list) -> Counts:
    """
    from news-search-api/client.py EsClientWrapper.format_count

    used to format "topdomains" & "toplangs" aggregation results

    takes [{"key": key, "doc_count": doc_count}, ....]
    and returns {key: count, ....}
    """
    return {item["key"]: item["doc_count"] for item in bucket}

def _b64_encode_page_token(strng: str) -> str:
    return base64.b64encode(strng.encode(), b"-_").decode().replace("=", "~")

def _b64_decode_page_token(strng: str) -> str:
    return base64.b64decode(strng.replace("~", "=").encode(), b"-_").decode()

# Used to concatenate multiple sort keys (before b64 encoding) and split
# after b64 decode.  Must not appear in key values!  Can be multi-character
# string to lower likelihood of appearing (default keys are numeric).
_SORT_KEY_SEP = "\x01"

def _get_hits(res: Response) -> list[Hit]:
    """
    retrieve hits array from _search results.
    currently called after all _search() calls
    """
    # _search method will have already logged any failed shards.
    # Response.success() wants
    # `self._shards.total == self._shards.successful and not self.timed_out`
    if not res.success():
        # XXX raise an Exception with useful error message?
        logger.warn("res.success() is False!")
        return []

    try:
        return res.hits
    except AttributeError:
        # XXX raise an Exception with useful error message?
        logger.warn("failed to get res.hits!")
        return []

class OnlineNewsMediaCloudESProvider(OnlineNewsMediaCloudProvider):
    """
    version of MC Provider going direct to ES.

    Consolidates query formatting/creation previously spread
    across multiple files:

    * web-search/mcweb/backend/search/utils.py (url_search_strings)
    * this file (domain search string)
    * mc-providers/mc_providers/mediacloud.py (date ranges)
    * news-search-api/api.py (aggregation result handling)
    * news-search-api/client.py (DSL, including aggegations)

    NOTE!!! Uses elasticsearch-dsl library as much as possible (rather
    than hand-formatted JSON/dicts to allow maximum mypy type
    enforcement!!!  Passing raw JSON means ES may silently not do what
    you hoped/expected, or may cause ES runtime errors that could have
    been detected earlier.
    """

    BASE_URL = "http://ramos.angwin:9200,http://woodward.angwin:9200,http://bradley.angwin:9200"
    WORDS_SAMPLE = 5000

    def __init__(self, **kwargs: Any):
        """
        Supported kwargs:

        "profile": bool or str
            if True, request profiling data, and log total ES CPU usage
            CAN pass string (filename) here, but feeding all the
            resulting JSON files to es-tools/collapse-esperf.py for
            flamegraphing could get you a mish-mash of different
            queries' results.
        "software_id": str (may be displayed by "mc-es-top" as "opaque_id")
        "session_id": str (user/session id for routing/caching)
        """

        self._profile: str | bool = kwargs.pop("profile", False)
        self._profile_current_search: str | bool = False

        # total seconds from the last profiled query:
        self._last_elastic_ms = -1.0

        # after pop-ing any local-only args:
        super().__init__(**kwargs)

        eshosts = self._base_url.split(",") # comma separated list of http://SERVER:PORT

        # Retries without delay (never mind backoff!)
        # web-search creates new Provider for each API request,
        # so randomize the pool.

        # https://www.elastic.co/guide/en/elasticsearch/reference/current/api-conventions.html
        # says:
        #   The X-Opaque-Id header accepts any arbitrary
        #   value. However, we recommend you limit these values to a
        #   finite set, such as an ID per client. Don’t generate a
        #   unique X-Opaque-Id header for every request. Too many
        #   unique X-Opaque-Id values can prevent Elasticsearch from
        #   deduplicating warnings in the deprecation logs.
        # See session_id for per-user/instance identification.

        self._es = elasticsearch.Elasticsearch(eshosts,
                                               max_retries=3,
                                               opaque_id=self._software_id,
                                               request_timeout=self._timeout,
                                               randomize_nodes_in_pool=True)


    def get_client(self):
        """
        called from OnlineNewsAbstractProvider constructor to set _client
        """
        return None

    def _fields(self, expanded: bool) -> list[Field]:
        """
        from news-search-api/client.py QueryBuilder constructor:
        return list of fields for item, paged_items, all_items to return

        NOTE: In the case of "indexed_date" the mapped field is currently stored
        with only millisecond resolution, while the stored string usually
        has microsecond resolution.
        """
        fields: list[Field] = [
            "article_title", "publication_date", "indexed_date",
            "language", "full_language", "canonical_domain", "url", "original_url"
        ]
        if expanded:
            fields.append("text_content")
        return fields

    # Allow weighting in order to (experimentally) apply filters in
    # most efficient order (most selective filter first).  An average
    # day is about 500K stories for all sources.  1K stories/day would
    # be a large source.  BUT adding a day means only expanding a
    # range, not adding another term...
    SELECTOR_WEIGHT = 1         # domains, filters, url_search_strings
    DAY_WEIGHT = 1

    @classmethod
    def _selector_filter_tuple(cls, kwargs: dict) -> FilterTuple:
        """
        function to allow construction of DSL
        rather than restorting to formatting/quoting query-string
        only to have ES have to parse it??
        For canonical_domain: "Match" query defaults to OR for space separated words
        For url: use "Wildcard"??
        Should initially take temp kwarg bool to allow A/B testing!!!
        """

        selector_clauses = cls._selector_query_clauses(kwargs)
        if selector_clauses:
            sqs = cls._selector_query_string_from_clauses(selector_clauses)
            return (cls._selector_count(kwargs) * cls.SELECTOR_WEIGHT,
                    SanitizedQueryString(query=sqs))
        else:
            return (0, None)

    def _basic_search(self, user_query: str, start_date: dt.datetime, end_date: dt.datetime,
                     expanded: bool = False, source: bool = True, **kwargs: Any) -> Search:
        """
        from news-search-api/api.py cs_basic_query
        create a elasticsearch_dsl query from user_query, date range, and kwargs
        """
        # works for date or datetime! publication_date is just YYYY-MM-DD
        start = start_date.strftime("%Y-%m-%d")
        end = end_date.strftime("%Y-%m-%d")

        self._profile_current_search = kwargs.pop("profile", self._profile)

        # check for extraneous arguments
        self._check_kwargs(kwargs)

        s = Search(index=self._index_from_dates(start_date, end_date), using=self._es)

        if self._profile_current_search:
            s = s.extra(profile=True)

        if user_query.strip() != self.everything_query(): # not "*"?
            s = s.query(SanitizedQueryString(query=user_query, default_field="text_content", default_operator="and"))

        if self._session_id:
            # pass user-id and/or session
            #   id to maximize ES caching effectiveness.
            # https://www.elastic.co/guide/en/elasticsearch/reference/7.17/search-search.html#search-preference
            #   If the cluster state and selected shards do not
            #   change, searches using the same <custom-string> value
            #   (that does not start with "_") are routed to the same
            #   shards in the same order.
            s = s.params(preference=self._session_id)

        # Evaluating selectors (domains/filters/url_search_strings) in "filter context";
        # Supposed to be faster, and enable caching of document set.
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/query-filter-context.html#filter-context

        # Try to apply filter with the smallest result set (most selective) first,
        # to cut down document set as soon as possible.

        days = (end_date - start_date).days + 1
        filters : list[FilterTuple] = [
            (days * self.DAY_WEIGHT, Range(publication_date={'gte': start, "lte": end})),
            self._selector_filter_tuple(kwargs)
            # could include languages (etc) here
        ]

        # try applying more selective queries (fewer results) first
        filters.sort()
        for weight, query in filters:
            if query:
                # ends up as list under bool.filter:
                s = s.filter(query)

        if source:              # return source (fields)?
            return s.source(self._fields(expanded))
        else:
            return s.source(False) # no source fields in hits

    def __repr__(self) -> str:
        return "OnlineNewsMediaCloudESProvider"

    def _is_no_results(self, results: Overview) -> bool:
        """
        used to test _overview_query results
        """
        return results is _EMPTY_OVERVIEW

    def _index_from_dates(self, start_date: dt.datetime | None, end_date: dt.datetime | None) -> list[str]:
        """
        return list of indices to search for a given date range.
        if indexing goes back to being split by publication_date (by year or quarter?)
        this could limit the number of shards that need to be queried
        """
        return [self._index]

    def _search(self, search: Search) -> Response:
        """
        one place to send queries to ES, for logging
        """
        logger.debug("MC._search %r", search.to_dict())

        t0 = time.monotonic()
        execute_args = {}
        if self._caching < 0:
            # Here to try to force ES not to use cached results (for testing).
            # Only effects in-library caching:
            execute_args["ignore_cache"] = True

            # This puts ?request_cache=false on the request URL, which
            # https://www.elastic.co/guide/en/elasticsearch/reference/current/shard-request-cache.html
            # says "The request_cache query-string parameter can be
            # used to enable or disable caching on a per-request
            # basis. If set, it overrides the index-level setting"
            search = search.params(request_cache=False)

        res = search.execute(**execute_args)
        elapsed = time.monotonic() - t0
        logger.info("MC._search ES took %s ms (%.3f elapsed)",
                    getattr(res, "took", -1), elapsed*1000)

        if (pdata := getattr(res, "profile", None)):
            self._process_profile_data(pdata)  # displays ES total time

        try:                    # look for circuit breaker trips, etc
            shards = res._shards
            if shards:
                failed = shards.failed
                total = shards.total
                if failed:
                    # hundreds of shards, so summarize...
                    # (almost always circuit breakers)
                    reasons: Counter[str] = Counter()
                    for shard in shards.failures:
                        rt = shard.reason.type
                        if rt:
                            reasons[rt] += 1
                    logger.info("MC._search %d/%d shards failed; reasons: %r",
                                failed, total, dict(reasons))
        except (ValueError, KeyError) as e:
            logger.debug("error looking at results: %r", e)
            # XXX raise Exception here??
        return res

    def _search_hits(self, search: Search) -> list[Hit]:
        """
        perform search, return list of Hit
        """
        res = self._search(search)
        return _get_hits(res)

    def _process_profile_data(self, pdata: AttrDict) -> None:
        """
        digest profiling data
        """
        pcs = self._profile_current_search # saved by _basic_search
        if isinstance(pcs, str):  # filename prefix?
            fname = time.strftime(f"{pcs}-%Y-%m-%d-%H-%M-%S.json")
            with open(fname, "w") as f:
                json.dump(pdata.to_dict(), f)
            logger.info("wrote profiling data to %s", fname)

        # sum up ES internal times
        query_ns = rewrite_ns = coll_ns = agg_ns = 0
        for shard in pdata.shards: # AttrList
            for search in shard.searches: # AttrList
                for q in search.query:    # AttrList
                    query_ns += q.time_in_nanos
                for coll in search.collector: # list
                    coll_ns += coll.time_in_nanos
                rewrite_ns += search.rewrite_time
            # XXX sum by aggregation name?
            for agg in shard.aggregations:
                agg_ns += agg.time_in_nanos
        es_nanos = query_ns + rewrite_ns + coll_ns + agg_ns
        self._last_elastic_ms = es_nanos / 1e6 # convert ns to ms
        logger.info("ES time: %.3f ms", self._last_elastic_ms)

        # avoid floating point divisions that are likely not displayed:
        # XXX save components???
        logger.debug(" ES (ns) query: %d rewrite: %d, collectors: %d aggs: %d",
                     query_ns, rewrite_ns, coll_ns, agg_ns)

    @CachingManager.cache('overview')
    def _overview_query(self, q: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs: Any) -> Overview:
        """
        from news-search-api/api.py
        returns _EMPTY_OVERVIEW when no hits
        """

        logger.debug("MC._overview %s %s %s %r", q, start_date, end_date, kwargs)

        # these are arbitrary, but match news-search-api/client.py
        # so that es-tools/mc-es-top.py can recognize this is an overview query:
        AGG_DAILY = "dailycounts"
        AGG_LANG = "toplangs"
        AGG_DOMAIN = "topdomains"

        search = self._basic_search(q, start_date, end_date, **kwargs)
        search.aggs.bucket(AGG_DAILY, "date_histogram", field="publication_date",
                           calendar_interval="day", min_doc_count=1)
        search.aggs.bucket(AGG_LANG, "terms", field="language.keyword", size=100)
        search.aggs.bucket(AGG_DOMAIN, "terms", field="canonical_domain", size=100)
        search = search.extra(track_total_hits=True)
        res = self._search(search) # run search
        hits = _get_hits(res)   # get hits list
        if not hits:
            return _EMPTY_OVERVIEW # checked by _is_no_results

        # res.hits.total.value documented at
        # https://elasticsearch-dsl.readthedocs.io/en/stable/search_dsl.html#response
        total = res.hits.total.value # type: ignore[attr-defined]
        aggs = res.aggregations

        return Overview(
            query=q,
            total=total,
            topdomains=_format_counts(aggs[AGG_DOMAIN]["buckets"]),
            toplangs=_format_counts(aggs[AGG_LANG]["buckets"]),
            dailycounts=_format_day_counts(aggs[AGG_DAILY]["buckets"]),
            matches=[_format_match(h) for h in hits], # _match_to_row called in .sample()
        )

    @CachingManager.cache()
    def item(self, item_id: str) -> Item:
        expanded = True         # always includes full_text!!
        s = Search(index=self._index, using=self._es)\
            .query(Match(_id=item_id))\
            .source(includes=self._fields(expanded)) 
        hits = self._search_hits(s)
        if not hits:
            return {}

        # double conversion!
        return self._match_to_row(_format_match(hits[0], expanded))

    def paged_items(
            self, query: str,
            start_date: dt.datetime, end_date: dt.datetime,
            page_size: int = 1000,
            **kwargs: Any
    ) -> tuple[list[dict], Optional[str]]:
        """
        return a single page of data (with `page_size` items).
        Pass `None` as first `pagination_token`, after that pass
        value returned by previous call, until `None` returned.

        `kwargs` may contain: `sort_field` (str), `sort_order` (str)
        """
        logger.debug("MC._paged_articles q: %s: %s e: %s ps: %d kw: %r",
                     query, start_date, end_date, page_size, kwargs)

        page_size = min(page_size, _ES_MAXPAGE)
        expanded = kwargs.pop("expanded", False)
        sort_field = kwargs.pop("sort_field", _DEF_SORT_FIELD)
        sort_order = kwargs.pop("sort_order", _DEF_SORT_ORDER)
        pagination_token = kwargs.pop("pagination_token", None)

        # NOTE! depends on client limiting to reasonable choices!!
        # (full text might leak data, or causes memory exhaustion!)
        if sort_field not in self._fields(expanded):
            raise ValueError(sort_field)
        if sort_order not in ["asc", "desc"]:
            raise ValueError(sort_order)

        # see discussion above at _SECONDARY_SORT_ARGS declaration
        sort_opts = [
            {sort_field: sort_order},
            _SECONDARY_SORT_ARGS
        ]

        search = self._basic_search(query, start_date, end_date, expanded=expanded, **kwargs)\
                     .extra(size=page_size, track_total_hits=True)\
                     .sort(*sort_opts)

        if pagination_token:
            # may return multiple keys:
            after = _b64_decode_page_token(pagination_token).split(_SORT_KEY_SEP)

            # important to use `search_after` instead of 'from' for
            # memory reasons related to paging through more than 10k
            # results.
            search = search.extra(search_after=after)

        hits = self._search_hits(search)
        if not hits:
            return ([], "")

        if len(hits) == page_size:
            # generate paging token from all sort keys of last item:
            new_pt = _b64_encode_page_token(
                _SORT_KEY_SEP.join([str(key) for key in hits[-1].meta.sort]))
        else:
            new_pt = ""

        # double conversion!
        rows = self._matches_to_rows([_format_match(h, expanded) for h in hits])
        return (rows, new_pt)

    def all_items(self, query: str,
                  start_date: dt.datetime, end_date: dt.datetime,
                  page_size: int = _ES_MAXPAGE, **kwargs: Any) -> AllItems:
        """
        returns generator of pages (lists) of items
        """
        next_page_token: str | None = None
        while True:
            page, next_page_token = self.paged_items(
                query, start_date, end_date,
                page_size=page_size,
                pagination_token=next_page_token,
                **kwargs)

            if not page:
                break

            yield page

            if not next_page_token:
                break

    def _sample_titles(self, query: str, start_date: dt.datetime, end_date: dt.datetime, sample_size: int,
                             **kwargs: Any) -> Iterable[list[dict[str,str]]]:
        """
        helper for generic Provider._sampled_title_words;
        returns a single page with entire sample_size of dicts
        """
        search = self._basic_search(query, start_date, end_date, **kwargs)\
                     .query(
                         FunctionScore(
                             # elasticsearch_dsl v8.17 gives mypy error, bug reported as
                             # https://github.com/elastic/elasticsearch-dsl-py/issues/1369
                             # PLEASE REMOVE THIS COMMENT WHEN THE IGNORE BELOW BECOMES UNNECESSARY!
                             functions=[ # type: ignore[arg-type]
                                 RandomScore(
                                     # needed for 100% reproducibility (ie; if paging results)
                                     # seed=int, field="fieldname"
                                 )
                             ]
                         )
                     )\
                     .source(["article_title", "language"])\
                     .extra(size=sample_size) # everything in one query

        hits = self._search_hits(search)
        ret = [{"title": hit.article_title, "language": hit.language} for hit in hits]
        return [ret]            # single page

    def words(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 100,
                             **kwargs: Any) -> list[Term]:
        """
        uses generic Provider._sampled_title_words
        with data from local helper above
        """
        return self._sampled_title_words(query, start_date, end_date, limit=limit, **kwargs)



import warnings


def deprecated(func):
    # This is a decorator which can be used to mark functions as deprecated. It will result in a
    # warning being emitted when the function is used.
    def new_func(*args, **kwargs):
        warnings.warn(f"Call to deprecated function {func.__name__}.", category=DeprecationWarning)
        return func(*args, **kwargs)
    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    new_func.__dict__.update(func.__dict__)
    return new_func




class ProviderException(Exception):
    pass


class UnsupportedOperationException(ProviderException):
    pass


class UnknownProviderException(ProviderException):
    def __init__(self, platform: str, source: str):
        super().__init__("Unknown provider {} from {}".format(platform, source))


class MissingRequiredValue(ProviderException):
    def __init__(self, name: str, keyword: str):
        super().__init__(f"provider {name} requires {keyword}")


class QueryingEverythingUnsupportedQuery(ProviderException):
    def __init__(self):
        super().__init__("Can't query everything")


# backwards compatibility:
APIKeyRequired = MissingRequiredValue
UnavailableProviderException = UnknownProviderException



import unittest
import datetime as dt
import pytest
import os
from mc_providers import (provider_for, PLATFORM_ONLINE_NEWS, PLATFORM_SOURCE_MEDIA_CLOUD)
from mc_providers.cache import CachingManager

MEDIA_CLOUD_API_KEY = os.getenv('MEDIA_CLOUD_API_KEY', None)

IN_GITHUB_CI_WORKFLOW = os.getenv("GITHUB_ACTIONS") == "true"

cache = {}


def _cache_get_key(*args, **kwargs):
    import hashlib
    serialise = []
    for arg in args:
        serialise.append(str(arg))
    for key, arg in kwargs.items():
        serialise.append(str(key))
        serialise.append(str(arg))
    key = hashlib.md5("".join(serialise).encode("UTF8")).hexdigest()
    return key


def _cache_handler(fn, cache_prefix, *args, **kwargs) -> tuple:
    was_cached = False
    key = _cache_get_key(cache_prefix, *args, **kwargs)
    results = cache.get(key)
    if not results:
        was_cached = False
        results = fn(*args, **kwargs)
        cache[key] = results
    return results, was_cached


@pytest.mark.skipif(IN_GITHUB_CI_WORKFLOW, reason="requires VPN tunnel to Media Cloud News Search API server")
class CacheMediaCloudTest(unittest.TestCase):

    def setUp(self) -> None:
        self._provider = provider_for(PLATFORM_ONLINE_NEWS, PLATFORM_SOURCE_MEDIA_CLOUD, None,
                                      "http://localhost:8010/v1/")
        CachingManager.cache_function = _cache_handler

    def _reset_cache(self):
        global cache
        cache = {}
        assert len(cache.keys()) == 0

    def _verify_cache_size(self, expected_size):
        assert len(cache.keys()) == expected_size

    def tearDown(self):
        self._reset_cache()
        CachingManager.cache_function = None

    def test_count(self):
        self._reset_cache()
        self._verify_cache_size(0)
        domains = ['usatoday.com', 'latimes.com', 'nypost.com']
        # prime cache
        results1 = self._provider.count("coronavirus", dt.datetime(2023, 11, 1),
                                        dt.datetime(2023, 12, 1),
                                        domains=domains)
        self._verify_cache_size(1)
        # verify first item cached
        results2 = self._provider.count("coronavirus", dt.datetime(2023, 11, 1),
                                        dt.datetime(2023, 12, 1),
                                        domains=domains)
        self._verify_cache_size(1)
        assert results1 == results2
        # add second
        results3 = self._provider.count("coronavirus", dt.datetime(2023, 11, 1),
                                        dt.datetime(2023, 12, 1))
        self._verify_cache_size(2)
        assert results1 != results3
        # verify second cached
        results4 = self._provider.count("coronavirus", dt.datetime(2023, 11, 1),
                                        dt.datetime(2023, 12, 1))
        self._verify_cache_size(2)
        assert results3 == results4
        # verify other query, which also uses `overview` under the hood, is a cache hit
        results5 = self._provider.sample("coronavirus", dt.datetime(2023, 11, 1),
                                         dt.datetime(2023, 12, 1))
        self._verify_cache_size(2)



import datetime as dt
import requests
from typing import List, Dict
import logging
import random
from collections import Counter
import numpy as np 
from time import sleep

from .errors import deprecated
from .provider import ContentProvider
from .exceptions import UnsupportedOperationException
from .cache import CachingManager
from .language import top_detected

TWITTER_DATE_FORMAT = '%Y-%m-%dT%H:%M:%S.%f%z'


@deprecated
class TwitterTwitterProvider(ContentProvider):
    """
    All these endpoints accept a `usernames: List[str]` keyword arg.
    """
    
    BASE_URL = 'https://api.twitter.com/2/'
    MAX_QUERY_LENGTH = 1024  # I think?
    POLITENESS_DELAY = 1  # sleep for half a second if we're gonna spam a bunch of queries

    def __init__(self, **kwargs):
        # used to take bearer_token: now use api_key
        super(TwitterTwitterProvider, self).__init__(**kwargs)
        self._session = requests.Session()  # better performance to put all HTTP through this one object

    #Chunk'd
    def sample(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 10, **kwargs) -> List[Dict]:
        """
        Return a list of historical tweets matching the query.
        :param query:
        :param start_date:
        :param end_date:
        :param limit:
        :param kwargs:
        :return:
        """
        # sample of historical tweets
        all_results = []
        chunks = self._assemble_and_chunk_query_str(query, **kwargs)
        for subquery in chunks:
            params = {
                "query": subquery,
                "max_results": limit,
                "start_time": start_date.isoformat("T") + "Z",
                "end_time": self._fix_end_date(end_date).isoformat("T") + "Z",
                "tweet.fields": ",".join(["author_id", "created_at", "public_metrics"]),
                "expansions": "author_id",
            }
            results = self._cached_query("tweets/search/all", params)
            results = TwitterTwitterProvider._tweets_to_rows(results)
            all_results.extend(results)
            if len(chunks) > 1:
                
                sleep(self.POLITENESS_DELAY)
                
        if len(all_results) > limit:
            all_results = random.sample(all_results, limit) 
        return results

    @classmethod
    def _assembled_query_str(cls, query: str, **kwargs) -> str:
        usernames = kwargs.get('usernames', [])
        # need to put all those filters in single query string
        if len(usernames) == 0:
            assembled_query = query
        else:
            assembled_query = query + " (" + " OR ".join(["from:{}".format(name) for name in usernames]) + ")"
        # check if query too long
        # @see https://developer.twitter.com/en/docs/twitter-api/tweets/search/api-reference/get-tweets-search-all
        # if len(assembled_query) > 1024:
        #    raise RuntimeError("Twitter's max query length is 1024 characters - your query is {} characters. "
        #                       "Try changing collections.".format(len(assembled_query)))
        return assembled_query

    
    @classmethod
    def _assemble_and_chunk_query_str(cls, base_query: str, **kwargs) -> List:
        """
        If a query string is too long, we can attempt to run it anyway by splitting the domain substring (which is
        guaranteed to be only a sequence of ANDs) into parts, to produce multiple smaller queries which are
        collectively equivalent to the original.
        """
        usernames = kwargs.get('usernames', [])
        
        if len(base_query) > cls.MAX_QUERY_LENGTH:
            # of course there still is the possibility that the base query is too large, which
            # cannot be fixed by this method
            raise RuntimeError(f"Base Query cannot exceed {cls.MAX_QUERY_LENGTH} characters")
        
        queries = [cls._assembled_query_str(base_query, usernames=usernames)]
        queries_too_big = any([len(q_) > cls.MAX_QUERY_LENGTH for q_ in queries])
        username_divisor = 2
        
        if queries_too_big:
            while queries_too_big:
                chunked_users = np.array_split(usernames, username_divisor)
                queries = [cls._assembled_query_str(base_query, usernames=users) for users in chunked_users]
                queries_too_big = any([len(q_) > cls.MAX_QUERY_LENGTH for q_ in queries])
                username_divisor *= 2
            
        return queries
    
    # No need to chunk
    def count(self, query: str, start_date: dt.datetime, end_date: dt.datetime, **kwargs) -> int:
        results = self.count_over_time(query, start_date, end_date, **kwargs)  # use the cached counts being made already
        total = sum([r['count'] for r in results['counts']])
        return total
    
    # Chunk
    def count_over_time(self, query: str, start_date: dt.datetime,
                        end_date: dt.datetime,
                        **kwargs) -> Dict:
        """
        Count how many tweets match the query over the 31 days before end_date.
        :param query:
        :param start_date:
        :param end_date:
        :param kwargs:
        :return:
        """
        counter: Counter = Counter()

        for subquery in  self._assemble_and_chunk_query_str(query, **kwargs):
            #print(subquery)
            params = dict(
                query=subquery,
                granularity='day',
                start_time=start_date.isoformat("T") + "Z",
                end_time=self._fix_end_date(end_date).isoformat("T") + "Z",
            )
            more_data = True
            next_token = None
            data = []
            while more_data:
                params['next_token'] = next_token
                results = self._cached_query("tweets/counts/all", params)
                data += reversed(results['data'])
                if ('meta' in results) and ('next_token' in results['meta']):
                    next_token = results['meta']['next_token']
                    more_data = True
                else:
                    next_token = None
                    more_data = False
            countable = {i["start"]:i["tweet_count"] for i in data}
            counter += Counter(countable)
                
        data = dict(counter)
        
        to_return = []
        for date, count in data.items():
            to_return.append({
                'date': dt.datetime.strptime(date, TWITTER_DATE_FORMAT),
                'timestamp': dt.datetime.strptime(date, TWITTER_DATE_FORMAT).timestamp(),
                'count': count,
            })
        return {'counts': to_return}

    # Chunk'd
    def all_items(self, query: str, start_date: dt.datetime, end_date: dt.datetime, page_size: int = 500,
                  **kwargs):
        
        for subquery in self._assemble_and_chunk_query_str(query, **kwargs):
            
            limit = kwargs['limit'] if 'limit' in kwargs else None
            next_token = None
            more_data = True
            params = {
                "query": subquery,
                "max_results": page_size,
                "start_time": start_date.isoformat("T") + "Z",
                "end_time": self._fix_end_date(end_date).isoformat("T") + "Z",
                "tweet.fields": ",".join(["author_id", "created_at", "public_metrics"]),
                "expansions": "author_id",
            }
            item_count = 0
            while more_data and ((limit is not None) and (item_count < limit)):
                params['next_token'] = next_token
                results = self._cached_query("tweets/search/all", params)
                result_count = results['meta']['result_count']
                if result_count == 0:
                    more_data = False
                    continue
                page = TwitterTwitterProvider._tweets_to_rows(results)
                yield page
                item_count += len(page)
                next_token = results['meta']['next_token'] if 'next_token' in results['meta'] else None
                more_data = next_token is not None
            
            sleep(self.POLITENESS_DELAY)
    
    # sampled_languages just relies on all_items, so if all_items is chunked then we get this for free
    def languages(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 10, **kwargs) -> List[Dict]:
        # use the helper because we need to sample from most recent tweets
        return self._sampled_languages(query, start_date, end_date, limit, **kwargs)
    
    # same as languages
    def words(self, query: str, start_date: dt.datetime, end_date: dt.datetime, limit: int = 100,
              **kwargs) -> List[Dict]:
        # use the helper because we need to sample from most recent tweets
        return self._sampled_title_words(query, start_date, end_date, limit, **kwargs)

    @CachingManager.cache()
    def _cached_query(self, endpoint: str, params: Dict = None) -> Dict:
        """
        Run a generic query agains the Twitter historical search API
        :param endpoint:
        :param query:
        :param params:
        :return:
        """
        headers = {
            'Content-type': 'application/json',
            'Authorization': "Bearer {}".format(self._api_key)
            # could use self._software_id for User-Agent!
        }
        r = self._session.get(self._base_url+endpoint, headers=headers, params=params, timeout=self._timeout)
        if r.status_code != 200:
            try:
                raise RuntimeError(r.json()['title'])
            except:
                raise RuntimeError(f"Error code {r.status_code}")
        return r.json()

    @classmethod
    def _add_author_to_tweets(cls, results: Dict) -> None:
        if cls._no_results(results):
            return
        user_id_lookup = {u['id']: u for u in results['includes']['users']}
        for t in results['data']:
            t['author'] = user_id_lookup[t['author_id']]

    @classmethod
    def _tweets_to_rows(cls, results: Dict) -> List:
        TwitterTwitterProvider._add_author_to_tweets(results)
        if cls._no_results(results):
            return []
        return [TwitterTwitterProvider._tweet_to_row(t) for t in results['data']]


    @classmethod
    def _no_results(self, results):
        return results['meta']['result_count'] == 0

    @classmethod
    def _tweet_to_row(cls, item: Dict) -> Dict:
        link = 'https://twitter.com/{}/status/{}'.format(item['author']['username'], item['id'])
        return {
            'media_name': '@'+item['author']['username'],
            'media_url': 'https://twitter.com/{}'.format(item['author']['username']),
            'id': item['id'],
            'title': item['text'],
            'publish_date': dt.datetime.strptime(item['created_at'], TWITTER_DATE_FORMAT),
            'url': link,
            'last_updated': dt.datetime.strptime(item['created_at'], TWITTER_DATE_FORMAT),
            'author': item['author']['name'],
            'language': top_detected(item['text']),  # guess the language cause Twitter oddly doesn't
            'retweet_count': item['public_metrics']['retweet_count'],
            'reply_count': item['public_metrics']['reply_count'],
            'like_count': item['public_metrics']['like_count'],
            'quote_count': item['public_metrics']['quote_count'],
            'content': item['text']
        }

    def normalized_count_over_time(self, query: str, start_date: dt.datetime, end_date: dt.datetime,
                                   **kwargs) -> Dict:
        raise UnsupportedOperationException("Can't search twitter for all tweets in a timeframe")

    def __repr__(self):
        # important to keep this unique among platforms so that the caching works right
        return "TwitterTwitterProvider"

    @classmethod
    def _fix_end_date(cls, orig_end_date: dt.datetime) -> dt.datetime:
        # twitter end dates are NOT inclusive, so we need to add one day here to make it match the general
        # behavior of our system (and UI copywriting)
        # also: can't search results more recent than 10 seconds ago (or something like that)
        return min(orig_end_date + dt.timedelta(days=1),
                   dt.datetime.now() - dt.timedelta(minutes=5))



from typing import Callable, Iterable, Optional

from .provider import ContentProvider

class CachingManager():
    """
    Use this to set caching functionality on the providers library for whatever system your using
    something like:
    `CachingManager.cache_function = your_cache_function`
    your_cache_function should have a signature like (function_to_cache, cache_prefix, args, kwargs) -> tuple[Any, bool]
    (second item in tuple indicatin if it was cached or not already)
    """
    cache_function = None

    @classmethod
    def cache(cls, custom_prefix_for_key: Optional[str] = None, kwargs_to_ignore: Iterable[str] = []):
        """
        @param custom_prefix_for_key: if specified, will be used in place of function name for cache_key generation
        """
        def decorator(fn: Callable):
            # WISH: detect if 'fn' is being declared as a method to a ContentProvider!!
            def wrapper(*args, **kwargs):
                # blindly assume that everything decorated is a ContentProvider method!
                inst = args[0]
                assert isinstance(inst, ContentProvider)
                # check caching enabled and wasn't disabled when Provider instantiated
                if cls.cache_function is not None and inst._caching > 0:
                    cache_prefix = custom_prefix_for_key or fn.__name__
                    # remove any kwargs already processed
                    # and included in positional args (ie; query string)
                    if kwargs_to_ignore and kwargs:
                        kwargs = kwargs.copy()
                        for kw in kwargs_to_ignore:
                            if kw in kwargs:
                                kwargs.pop(kw)
                    results, was_cached = cls.cache_function(fn, cache_prefix, *args, **kwargs)
                    return results
                else:
                    return fn(*args, **kwargs)
            return wrapper
        return decorator



# pylint: disable=protected-access

import unittest
import datetime as dt
import os

from ..youtube import YouTubeYouTubeProvider
from ..exceptions import UnsupportedOperationException

YOUTUBE_API_KEY = os.getenv('YOUTUBE_API_KEY', None)

TERM = "robot"
DAY_WINDOW = 100  # window in days to search for tem


class YouTubeYouTubeProviderTest(unittest.TestCase):

    def setUp(self):
        self._provider = YouTubeYouTubeProvider(YOUTUBE_API_KEY)

    def test_fetch_results_from_api(self):
        results = self._provider._fetch_results_from_api(
            TERM,
            start_date=dt.datetime.now() - dt.timedelta(days=DAY_WINDOW),
            end_date=dt.datetime.now()
        )
        assert isinstance(results, dict) is True

    def test_count(self):
        try:
            _ = self._provider.count(
                TERM,
                start_date=dt.datetime.now() - dt.timedelta(days=2),
                end_date=dt.datetime.now()
            )
        except UnsupportedOperationException:
            assert True

    def test_count_too_big(self):
        try:
            _ = self._provider.count(
                TERM,
                start_date=dt.datetime.now() - dt.timedelta(days=DAY_WINDOW),
                end_date=dt.datetime.now()
            )
        except UnsupportedOperationException:
            assert True

    def test_sample(self):
        results = self._provider.sample(
            TERM,
            start_date=dt.datetime.now() - dt.timedelta(days=DAY_WINDOW),
            end_date=dt.datetime.now()
        )
        assert isinstance(results, list) is True

    def test_all_items(self):
        # Something about this query is not deterministic- result ranges +-10 from 280, leading to failures
        page_count = 0
        total_items = 0
        for page in self._provider.all_items("cultural marxism", start_date=dt.datetime(2023, 1, 1),
                                             end_date=dt.datetime(2023, 1, 5)):
            assert len(page) > 0
            total_items += len(page)
            page_count += 1
        assert page_count == 1
        assert total_items == 2
        page_count = 0
        total_items = 0
        for page in self._provider.all_items(TERM, start_date=dt.datetime(2023, 1, 1),
                                             end_date=dt.datetime(2023, 1, 10)):
            assert len(page) > 0
            total_items += len(page)
            page_count += 1
        assert page_count > 1
        assert total_items > 0
        # assert abs(total_items - 280) < 10  #since there seems to be some fuzziness, and I want to supress the error.
